/*
  gsl_pd_histogram_distributions.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <vector>
#include <cstdlib>
#include <cassert>
#include <cmath>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_histogram.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sort.h>
#if __cplusplus < 201103L || defined(USE_BOOST_FUNCTION)
#include <boost/function.hpp>
#endif
#if __cplusplus < 201103L || defined(USE_BOOST_UTILITY)
#include <boost/utility.hpp>
#endif
#include "averages.hpp"
#include "statistics.hpp"
#include "histograms.hpp"
#include "histograms_online.hpp"

#define numberof(a) (sizeof(a)/sizeof(a[0]))

template <typename InputIterator>
std::vector<std::string> tabs_for_string_list(const InputIterator first, const InputIterator last, const ptrdiff_t offset = 0)
{
  std::vector<std::string> rv;
  InputIterator it(first);
  size_t l = std::string(*it++).size();
  while (it != last)
    l = std::max(l, std::string(*it++).size());
  for (it = first; it != last; ++it) {
    rv.push_back(std::string("\t"));
    for (size_t j=0; j<(l+offset)/8 - (std::string(*it).size()+offset)/8; ++j)
      rv[rv.size()-1] += "\t";
  }
  return rv;
}

double distribution_type_value(const size_t t, gsl_rng *r)
{
  double rv = 0;

  switch (t) {
  case 0:	rv = gsl_rng_uniform(r);	break;
  case 1:	rv = gsl_ran_ugaussian(r);	break;
  case 2:	rv = gsl_ran_gaussian(r, 1.);	break;
  case 3:	rv = gsl_ran_ugaussian_tail(r, 1.5);	break;
  case 4:	rv = gsl_ran_gaussian_tail(r, 1.5, 1.);	break;
  case 5:	rv = gsl_ran_exponential(r, 1.);	break;
  case 6:	rv = gsl_ran_laplace(r, 1.);	break;
  case 7:	rv = gsl_ran_exppow(r, 1., 2.5);	break;
  case 8:	rv = gsl_ran_cauchy(r, 1.);	break;
  case 9:	rv = gsl_ran_rayleigh(r, 1.);	break;
  case 10:	rv = gsl_ran_rayleigh_tail(r, 1., 1.);	break;
  case 11:	rv = gsl_ran_landau(r);	break;
  case 12:	rv = gsl_ran_levy(r, 1., 1.);	break;
  case 13:	rv = gsl_ran_levy_skew(r, 1., 1., 1.);	break;
  case 14:	rv = gsl_ran_gamma(r, 1., 1.);	break;
  case 15:	rv = gsl_ran_flat(r, 0., 1.);	break;
  case 16:	rv = gsl_ran_chisq(r, 1.);	break;
  case 17:	rv = gsl_ran_lognormal(r, 0., 1.);	break;
  case 18:	rv = gsl_ran_fdist(r, 3., 2.);	break;
  case 19:	rv = gsl_ran_tdist(r, 1.);	break;
  case 20:	rv = gsl_ran_beta(r, 2., 2.);	break;
  case 21:	rv = gsl_ran_logistic(r, 1.);	break;
  case 22:	rv = gsl_ran_pareto(r, 1., 1.);	break;
  case 23:	rv = gsl_ran_weibull(r, 1., 1.);	break;
  case 24:	rv = gsl_ran_gumbel1(r, 1., 1.);	break;
  case 25:	rv = gsl_ran_gumbel2(r, 1., 1.);	break;
  case 26:	rv = gsl_ran_poisson(r, 2.5);	break;
  case 27:	rv = gsl_ran_bernoulli(r, .7);	break;
  case 28:	rv = gsl_ran_binomial(r, .5, 9);	break;
  case 29:	rv = gsl_ran_negative_binomial(r, .5, 3.5);	break;
  case 30:	rv = gsl_ran_pascal(r, .5, 3);	break;
  case 31:	rv = gsl_ran_geometric(r, .5);	break;
  case 32:	rv = gsl_ran_hypergeometric(r, 5, 20, 3);	break;
  default:	rv = gsl_ran_logarithmic(r, .7);	break;
  }
  return rv;
}
const char *distribution_names[] = {
  "uniform",
  "ugaussian",
  "gaussian",
  "ugaussian_tail",
  "gaussian_tail",
  "exponential",
  "laplace",
  "exppow",
  "cauchy",
  "rayleigh",
  "rayleigh_tail",
  "landau",
  "levy",
  "levy_skew",
  "gamma",
  "flat",
  "chisq",
  "lognormal",
  "fdist",
  "tdist",
  "beta",
  "logistic",
  "pareto",
  "weibull",
  "gumbel1",
  "gumbel2",
  "poisson",
  "bernoulli",
  "binomial",
  "negative_binomial",
  "pascal",
  "geometric",
  "hypergeometric",
  "logarithmic",
};

int main()
{
  const double maximum_permissible_error(std::numeric_limits<double>::min());
  const double online_maximum_permissible_error(std::numeric_limits<double>::min());
  const double gsl_pd_maximum_permissible_error(
#if defined(HISTOGRAM_CUMULATIVE_DISTRIBUTION_GSL_COMPATIBLE) && \
    defined(HISTOGRAM_PSEUDO_DENSITY_GSL_COMPATIBLE)
						std::numeric_limits<double>::min()
#else
						//1e-10 // *_PSEUDO_DENSITY_GSL_COMPATIBLE
						1e-7 // *_CUMULATIVE_DISTRIBUTION_GSL_COMPATIBLE
#endif
						);
  const char *result_names[] = {
    "minimum value",
    "maximum value",
    "histogram min_element",
    "histogram min_value",
    "histogram max_element",
    "histogram max_value",
    "histogram mean",
    "histogram sigma",
    "histogram sum",
  };
  const size_t n(1000000), n_bins(10000), n_prior_points(n);
  double *a = (double *)malloc(sizeof(double)*n);
  const gsl_rng_type *T = gsl_rng_default;
  gsl_rng *r;

  std::vector<std::string> tabs_for_result_names(tabs_for_string_list(result_names, result_names + numberof(result_names), 1));

  gsl_rng_env_setup();
  r = gsl_rng_alloc(T);

  histogram_online<double> histogram_online(n_bins, n_prior_points),
    gsl_pseudo_dense_histogram_online(n_bins, n_prior_points),
    pseudo_dense_histogram_online(n_bins, n_prior_points);

  for (size_t t=0; t<33; ++t) {
    std::vector<double> v;
    std::cout << distribution_names[t] << ":" << std::endl;
    gsl_rng_set(r, gsl_rng_default_seed);
    for (size_t i=0; i<n; ++i) {
      a[i] = distribution_type_value(t, r);
      histogram_online.add(a[i]);
      v.push_back(a[i]);
    }

    double gsl_result_range[] = {
      gsl_stats_min(a, 1, n),
      gsl_stats_max(a, 1, n),
    };
    gsl_histogram *gsl_hst = gsl_histogram_alloc(n_bins);
    gsl_histogram_set_ranges_uniform(gsl_hst, gsl_result_range[0], gsl_result_range[1]);
    for (size_t i=0; i<n; ++i)
      gsl_histogram_increment(gsl_hst, a[i]);
    double gsl_results[] = {
      gsl_result_range[0],
      gsl_result_range[1],
      (double)gsl_histogram_min_bin(gsl_hst),
      gsl_histogram_min_val(gsl_hst),
      (double)gsl_histogram_max_bin(gsl_hst),
      gsl_histogram_max_val(gsl_hst),
      gsl_histogram_mean(gsl_hst),
      gsl_histogram_sigma(gsl_hst),
      gsl_histogram_sum(gsl_hst),
    };

    double result_range[] = {
      *min_element(v.begin(), v.end()),
      *max_element(v.begin(), v.end()),
    };
    std::vector<size_t> histogram_result(n_bins);
    size_t histogram_result_summation(histogram(histogram_result, v.begin(), v.end(), result_range[0], result_range[1]));
    double results[] = {
      result_range[0],
      result_range[1],
      (double)std::distance(histogram_result.begin(), min_element(histogram_result.begin(), histogram_result.end())),
      (double)*min_element(histogram_result.begin(), histogram_result.end()),
      (double)std::distance(histogram_result.begin(), max_element(histogram_result.begin(), histogram_result.end())),
      (double)*max_element(histogram_result.begin(), histogram_result.end()),
      histogram_mean(histogram_result, result_range[0], result_range[1]),
      histogram_standard_deviation(histogram_result, result_range[0], result_range[1]),
#if !0
      (double)histogram_result_summation,
#else
      (double)std::accumulate(histogram_result.begin(), histogram_result.end(), size_t(0)),
#endif
    };

    double online_results[] = {
      histogram_online.get_minimum_value(),
      histogram_online.get_maximum_value(),
      (double)histogram_online.get_minimum_height_index(),
      histogram_online.get_minimum_height(),
      (double)histogram_online.get_maximum_height_index(),
      histogram_online.get_maximum_height(),
      histogram_online.get_heights_mean(),
      histogram_online.get_heights_standard_deviation(),
      (double)histogram_online.get_heights_summation(),
    };

    std::cout << "gsl_results[i] - results[i]:" << std::endl;
    for (size_t j=0; j<numberof(results); ++j) {
      double error(sqrt(pow(gsl_results[j] - results[j], 2)) / (results[j] ? fabs(results[j]) : 1));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << gsl_results[j] << "\t" << results[j] << "\t" << error << std::endl;
      assert(error < maximum_permissible_error);
    }
    std::cout << std::endl;
    for (size_t b=0; b<n_bins; ++b) {
      if (!(gsl_hst->range[b] == histogram_range(b, n_bins, result_range[0], result_range[1]).first &&
	    gsl_hst->bin[b] == histogram_result[b]))
	std::cout << "gsl_hst->bin[] - histogram_result[]:\t" << b << "\t" <<
	  gsl_hst->range[b] << "\t" << gsl_hst->bin[b] << "\t" <<
	  histogram_range(b, n_bins, result_range[0], result_range[1]).first << "\t" << histogram_result[b] << std::endl;
      assert(gsl_hst->range[b] == histogram_range(b, n_bins, result_range[0], result_range[1]).first);
      assert(gsl_hst->bin[b] == histogram_result[b]);
    }

    std::cout << "online_results[i] - results[i]:" << std::endl;
    for (size_t j=0; j<numberof(results); ++j) {
      double error(sqrt(pow(online_results[j] - results[j], 2)) / (results[j] ? fabs(results[j]) : 1));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << online_results[j] << "\t" << results[j] << "\t" << error << std::endl;
      assert(error < online_maximum_permissible_error);
    }
    std::cout << std::endl;
    for (size_t b=0; b<n_bins; ++b) {
      if (!(histogram_online.get_lower_bound(b) == histogram_range(b, n_bins, result_range[0], result_range[1]).first &&
	    histogram_online.get_height(b) == histogram_result[b]))
	std::cout << "online_hst->bin[] - histogram_result[]:\t" << b << "\t" <<
	  histogram_online.get_lower_bound(b) << "\t" << histogram_online.get_height(b) << "\t" <<
	  histogram_range(b, n_bins, result_range[0], result_range[1]).first << "\t" << histogram_result[b] << std::endl;
      assert(histogram_online.get_lower_bound(b) == histogram_range(b, n_bins, result_range[0], result_range[1]).first);
      assert(histogram_online.get_height(b) == histogram_result[b]);
    }

#if !0 /* worse implement */
    std::vector<double> pd_histogram(histogram_cumulative_distribution<double>(histogram_result));
#else
    std::pair<std::vector<double>, std::vector<double> > pd_histogram(histogram_probability_cumulative_distribution<double>(histogram_result));
#endif

    gsl_histogram_pdf *gsl_hst_pdf = gsl_histogram_pdf_alloc(n_bins);
    gsl_histogram_pdf_init(gsl_hst_pdf, gsl_hst);

    gsl_rng_set(r, gsl_rng_default_seed);
    for (size_t i=0; i<n; ++i) {
      double x(gsl_rng_uniform(r));
      gsl_pseudo_dense_histogram_online.add(gsl_histogram_pdf_sample(gsl_hst_pdf, x));
#if !0 /* worse implement */
      pseudo_dense_histogram_online.add(histogram_cumulative_distribution_pseudo(pd_histogram, x,
										 histogram_result, result_range[0], result_range[1]));
#else
      pseudo_dense_histogram_online.add(histogram_probability_cumulative_distribution_pseudo(pd_histogram, x,
											     histogram_result, result_range[0], result_range[1]));
#endif
    }

    double gsl_pd_results[] = {
      gsl_pseudo_dense_histogram_online.get_minimum_value(),
      gsl_pseudo_dense_histogram_online.get_maximum_value(),
      (double)gsl_pseudo_dense_histogram_online.get_minimum_height_index(),
      gsl_pseudo_dense_histogram_online.get_minimum_height(),
      (double)gsl_pseudo_dense_histogram_online.get_maximum_height_index(),
      gsl_pseudo_dense_histogram_online.get_maximum_height(),
      gsl_pseudo_dense_histogram_online.get_heights_mean(),
      gsl_pseudo_dense_histogram_online.get_heights_standard_deviation(),
      (double)gsl_pseudo_dense_histogram_online.get_heights_summation(),
    };
    double pd_results[] = {
      pseudo_dense_histogram_online.get_minimum_value(),
      pseudo_dense_histogram_online.get_maximum_value(),
      (double)pseudo_dense_histogram_online.get_minimum_height_index(),
      pseudo_dense_histogram_online.get_minimum_height(),
      (double)pseudo_dense_histogram_online.get_maximum_height_index(),
      pseudo_dense_histogram_online.get_maximum_height(),
      pseudo_dense_histogram_online.get_heights_mean(),
      pseudo_dense_histogram_online.get_heights_standard_deviation(),
      (double)pseudo_dense_histogram_online.get_heights_summation(),
    };

    std::cout << "gsl_pd_results[i] - pd_results[i]:" << std::endl;
    for (size_t j=0; j<numberof(pd_results); ++j) {
      double error(sqrt(pow(gsl_pd_results[j] - pd_results[j], 2)) / (pd_results[j] ? fabs(pd_results[j]) : 1));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << gsl_pd_results[j] << "\t" << pd_results[j] << "\t" << error << std::endl;
      assert(error < gsl_pd_maximum_permissible_error);
    }
    std::cout << std::endl;
    
    histogram_online.reset();
    gsl_pseudo_dense_histogram_online.reset();
    pseudo_dense_histogram_online.reset();

    gsl_histogram_free(gsl_hst);
    gsl_histogram_pdf_free(gsl_hst_pdf);
  }
  gsl_rng_free(r);
  return 0;
}
