/*
  histogram.hpp

  Copyright (C) 2016 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _HISTOGRAM_HPP_
#define _HISTOGRAM_HPP_

#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
#define INTERNAL_VALUE_TYPE long double
#else
#define INTERNAL_VALUE_TYPE T
#endif

template <typename InputIterator>
size_t histogram(std::vector<size_t> &heights,
		 const InputIterator first, const InputIterator last,
		 const typename std::iterator_traits<InputIterator>::value_type lb,
		 const typename std::iterator_traits<InputIterator>::value_type ub,
		 const bool logarithmic = false)
{
  using std::log;
  size_t rv(0);
  const typename std::iterator_traits<InputIterator>::value_type sz(!logarithmic ? ub - lb : log(ub) - log(lb));
  for (InputIterator it(first); it!=last; ++it) {
    if (*it < lb) continue;
    const typename std::iterator_traits<InputIterator>::value_type v(!logarithmic ? *it - lb : log(*it) - log(lb));
    const size_t id(v/sz*heights.size());
    if (!(id < heights.size())) continue;
    ++heights[id];
    ++rv;
  }
  return rv;
}
template <typename T>
inline
T histogram_lower_bound(const size_t i, const size_t n_bins, const T lb, const T ub, const bool logarithmic = false, const bool option_lt_ub = false)
{
  using std::exp;
  using std::log;
#if 0
  if (option_lt_ub)
    return !logarithmic ? lb + (ub - lb)/n_bins*i : lb*exp((log(ub) - log(lb))*i/n_bins);
  else
    return !logarithmic ? T(n_bins - i)/n_bins*lb + T(i)/n_bins*ub : exp((T)(n_bins - i)/n_bins*log(lb) + (T)i/n_bins*log(ub));
#else
  if (option_lt_ub)
    return !logarithmic ? lb + (ub - lb)/n_bins*i : lb*exp((log((INTERNAL_VALUE_TYPE)ub) - log((INTERNAL_VALUE_TYPE)lb))*i/n_bins);
  else
    return !logarithmic ? T(n_bins - i)/n_bins*lb + T(i)/n_bins*ub : exp((INTERNAL_VALUE_TYPE)(n_bins - i)/n_bins*log((INTERNAL_VALUE_TYPE)lb) + (INTERNAL_VALUE_TYPE)i/n_bins*log((INTERNAL_VALUE_TYPE)ub));
#endif
}
template <typename T>
inline
T histogram_upper_bound(const size_t i, const size_t n_bins, const T lb, const T ub, const bool logarithmic = false, const bool option_lt_ub = false)
{
  return histogram_lower_bound(i+1, n_bins, lb, ub, logarithmic, option_lt_ub);
}
template <typename T>
inline
std::pair<T, T> histogram_range(const size_t i, const size_t n_bins, const T lb, const T ub, const bool logarithmic = false, const bool option_lt_ub = false)
{
  return std::make_pair(histogram_lower_bound(i, n_bins, lb, ub, logarithmic, option_lt_ub),
			histogram_upper_bound(i, n_bins, lb, ub, logarithmic, option_lt_ub));
}
template <typename T>
T histogram_mean(const std::vector<size_t> &heights, const T lb, const T ub, const bool logarithmic = false)
{
  INTERNAL_VALUE_TYPE w(0), v(0);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    if (!(*it > 0)) continue;
    const std::pair<T, T> range(histogram_range(std::distance(heights.begin(), it), heights.size(), lb, ub, logarithmic));
#if defined(HISTOGRAM_STATISTICS_NON_LOGARITHMIC_INTERPOLATION_GSL_COMPATIBLE)
    const INTERNAL_VALUE_TYPE p((range.first + range.second)/2);
#else
    const INTERNAL_VALUE_TYPE p(!logarithmic ? (range.first + range.second)/2 : exp((log(range.first) + log(range.second))/2));
#endif
    w += *it;
    v += (p - v)*((INTERNAL_VALUE_TYPE)*it/w);
  }
  return v;
}
template <typename T>
T histogram_variance(const std::vector<size_t> &heights, const T lb, const T ub, const bool logarithmic = false)
{
  using std::exp;
  using std::log;
  INTERNAL_VALUE_TYPE mu(histogram_mean(heights, lb, ub, logarithmic)), w(0), v(0);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    if (!(*it > 0)) continue;
    const std::pair<T, T> range(histogram_range(std::distance(heights.begin(), it), heights.size(), lb, ub, logarithmic));
#if defined(HISTOGRAM_STATISTICS_NON_LOGARITHMIC_INTERPOLATION_GSL_COMPATIBLE)
    const INTERNAL_VALUE_TYPE p((range.first + range.second)/2), delta(p - mu);
#else
    const INTERNAL_VALUE_TYPE p(!logarithmic ? (range.first + range.second)/2 : exp((log(range.first) + log(range.second))/2)), delta(p - mu);
#endif
    w += *it;
    v += (delta*delta - v)*((INTERNAL_VALUE_TYPE)*it/w);
  }
  return v;
}
template <typename T>
T histogram_standard_deviation(const std::vector<size_t> &heights, const T lb, const T ub, const bool logarithmic = false)
{
  using std::sqrt;
  return sqrt(histogram_variance(heights, lb, ub, logarithmic));
}

#if defined(HISTOGRAM_CUMULATIVE_DISTRIBUTION_GSL_COMPATIBLE)
template <typename T>
std::vector<T> histogram_cumulative_distribution(const std::vector<size_t> &heights)
{
  std::vector<T> cd(heights.size());
  T mu(0), sigma(0);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it)
    mu += (*it - mu)/(std::distance(heights.begin(), it) + 1);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    sigma += *it/mu/heights.size();
    cd[std::distance(heights.begin(), it)] = sigma;
  }
  return cd;
}
#else
template <typename T>
std::vector<T> histogram_cumulative_distribution(const std::vector<size_t> &heights)
{
  std::vector<T> cd(heights.size());
  T mu(0), sigma(0);
  size_t c(0);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    if (!*it) continue;
    mu += (*it - mu)/++c;
  }
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    const std::vector<size_t>::difference_type i = std::distance(heights.begin(), it);
    if (!*it) { cd[i] = sigma; continue; }
    sigma += *it/mu/c;
    cd[i] = sigma;
  }
  return cd;
}
#endif

template <typename T>
T histogram_cumulative_distribution_pseudo(const std::vector<T> &cd, const T r,
					   const std::vector<size_t> &heights, const T lb, const T ub, const bool logarithmic = false, const bool option_lt_ub = false)
{
  const typename std::vector<T>::const_iterator it = lower_bound(cd.begin(), cd.end(), r);
  const typename std::vector<T>::difference_type i = std::distance(cd.begin(), it);
  const std::pair<T, T> range(histogram_range(i, heights.size(), lb, ub, logarithmic, option_lt_ub));
  const T delta((r - (i==0 ? T(0) : cd[i-1]))/(cd[i] - (i==0 ? T(0) : cd[i-1])));
#if defined(HISTOGRAM_PSEUDO_DENSITY_NON_LOGARITHMIC_INTERPOLATION_GSL_COMPATIBLE)
  return
    range.first + delta*(range.second - range.first);
#else
  return !logarithmic ?
    range.first + delta*(range.second - range.first) :
    range.first*exp(delta*(log(range.second) - log(range.first)));
#endif
}

#if defined(HISTOGRAM_CUMULATIVE_DISTRIBUTION_GSL_COMPATIBLE)
template <typename T>
std::pair<std::vector<T>, std::vector<T> > histogram_probability_cumulative_distribution(const std::vector<size_t> &heights)
{
  std::vector<T> pd(heights.size()), cd(heights.size());
  T mu(0), sigma(0);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it)
    mu += (*it - mu)/(std::distance(heights.begin(), it) + 1);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    const std::vector<size_t>::difference_type i = std::distance(heights.begin(), it);
    pd[i] = *it/mu/heights.size();
    sigma += pd[i];
    cd[i] = sigma;
  }
  return std::make_pair(pd, cd);
}
#else
template <typename T>
std::pair<std::vector<T>, std::vector<T> > histogram_probability_cumulative_distribution(const std::vector<size_t> &heights)
{
  std::vector<T> pd(heights.size()), cd(heights.size());
  T mu(0), sigma(0);
  size_t c(0);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    if (!*it) continue;
    mu += (*it - mu)/++c;
  }
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    const std::vector<size_t>::difference_type i = std::distance(heights.begin(), it);
    if (!*it) { pd[i] = T(0); cd[i] = sigma; continue; }
    pd[i] = *it/mu/c;
    sigma += pd[i];
    cd[i] = sigma;
  }
  return std::make_pair(pd, cd);
}
#endif
template <typename T>
T histogram_probability_cumulative_distribution_pseudo(const std::pair<std::vector<T>, std::vector<T> > &pd_cd, const T r,
						       const std::vector<size_t> &heights, const T lb, const T ub, const bool logarithmic = false, const bool option_lt_ub = false)
{
  const typename std::vector<T>::const_iterator it = lower_bound(pd_cd.second.begin(), pd_cd.second.end(), r);
  const typename std::vector<T>::difference_type i = std::distance(pd_cd.second.begin(), it);
  const std::pair<T, T> range(histogram_range(i, heights.size(), lb, ub, logarithmic, option_lt_ub));
#if defined(HISTOGRAM_PSEUDO_DENSITY_GSL_COMPATIBLE)
  const T delta((r - (i==0 ? T(0) : pd_cd.second[i-1]))/(pd_cd.second[i] - (i==0 ? T(0) : pd_cd.second[i-1])));
#if defined(HISTOGRAM_PSEUDO_DENSITY_NON_LOGARITHMIC_INTERPOLATION_GSL_COMPATIBLE)
  return
    range.first + delta*(range.second - range.first);
#else
  return !logarithmic ?
    range.first + delta*(range.second - range.first) :
    range.first*exp(delta*(log(range.second) - log(range.first)));
#endif
#else
  const T delta((pd_cd.second[i] - r)/pd_cd.first[i]);
#if defined(HISTOGRAM_PSEUDO_DENSITY_NON_LOGARITHMIC_INTERPOLATION_GSL_COMPATIBLE)
  return
    delta*range.first + (1 - delta)*range.second;
#else
  return !logarithmic ?
    delta*range.first + (1 - delta)*range.second :
    exp(delta*log(range.first) + (1 - delta)*log(range.second));
#endif
#endif
}

#endif
