/*
  multidimensional_histogram.hpp

  Copyright (C) 2016 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _MULTIDIMENSIONAL_HISTOGRAM_HPP_
#define _MULTIDIMENSIONAL_HISTOGRAM_HPP_
#include "vector_numeric.hpp"
#include "vector_numeric_quantified_operators.hpp"
#include <map>
#include <numeric>
#if __cplusplus < 201103L || defined(USE_BOOST_UTILITY)
#if !defined(BOOST_UTILITY_HPP)
template <typename InputIterator>
inline InputIterator next(InputIterator it) { return ++it; }
template <typename InputIterator, typename Difference>
inline InputIterator next(InputIterator it, Difference n)
{
  std::advance(it, n);
  return it;
}
template <typename InputIterator>
inline InputIterator prior(InputIterator it) { return --it; }
template <typename InputIterator, typename Difference>
inline InputIterator prior(InputIterator it, Difference n)
{
  std::advance(it, -n);
  return it;
}
#define prev prior
#else
using boost::next;
using boost::prior;
#define prev prior
#endif
#else
using std::next;
using std::prev;
#endif

#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
#define INTERNAL_VALUE_TYPE long double
#else
#define INTERNAL_VALUE_TYPE T
#endif

template <typename InputIterator, typename SizeType, typename FlagType>
size_t histogram(std::map<std::vector<size_t>, size_t> &heights,
		 const InputIterator first, const InputIterator last,
		 const SizeType n_bins,
		 const std::vector<typename std::iterator_traits<InputIterator>::value_type::value_type> lb,
		 const std::vector<typename std::iterator_traits<InputIterator>::value_type::value_type> ub,
		 const FlagType &logarithmic)
{
  size_t rv(0);
  const std::vector<typename std::iterator_traits<InputIterator>::value_type::value_type> sz(operator_conditional(!logarithmic, ub - lb, log(ub) - log(lb)));
  for (InputIterator it(first); it!=last; ++it) {
    if (/**it < lb*/exists_operator_less(*it, lb)) continue;
    const std::vector<typename std::iterator_traits<InputIterator>::value_type::value_type> v(operator_conditional(!logarithmic, *it - lb, log(*it) - log(lb)));
    const std::vector<size_t> id(container_cast<std::vector<size_t> >(v/sz*n_bins));
    if (/*!(id < n_bins)*/!forall_operator_less(id, n_bins)) continue;
    ++heights[id];
    ++rv;
  }
  return rv;
}
template <typename T, typename SizeType, typename FlagType>
inline
std::vector<T> histogram_lower_bound(const std::vector<size_t> &i, const SizeType &n_bins, const std::vector<T> &lb, const std::vector<T> &ub, const FlagType &logarithmic, const bool option_lt_ub = false)
{
#if 0
  if (option_lt_ub)
    return operator_conditional(!logarithmic, lb + (ub - lb)/n_bins*i, lb*exp((log(ub) - log(lb))*i/n_bins));
  else
    return operator_conditional(!logarithmic, container_cast<std::vector<T> >(n_bins - i)/n_bins*lb + container_cast<std::vector<T> >(i)/n_bins*ub, exp(container_cast<std::vector<T> >(n_bins - i)/n_bins*log(lb) + container_cast<std::vector<T> >(i)/n_bins*log(ub)));
#else
  if (option_lt_ub)
    return operator_conditional(!logarithmic, lb + (ub - lb)/n_bins*i, lb*exp((log(container_cast<std::vector<INTERNAL_VALUE_TYPE> >(ub)) - log(container_cast<std::vector<INTERNAL_VALUE_TYPE> >(lb)))*i/n_bins));
  else
    return operator_conditional(!logarithmic, container_cast<std::vector<T> >(n_bins - i)/n_bins*lb + container_cast<std::vector<T> >(i)/n_bins*ub, container_cast<std::vector<T> >(exp(container_cast<std::vector<INTERNAL_VALUE_TYPE> >(n_bins - i)/n_bins*log(container_cast<std::vector<INTERNAL_VALUE_TYPE> >(lb)) + container_cast<std::vector<INTERNAL_VALUE_TYPE> >(i)/n_bins*log(container_cast<std::vector<INTERNAL_VALUE_TYPE> >(ub)))));
#endif
}
template <typename T, typename SizeType, typename FlagType>
inline
std::vector<T> histogram_upper_bound(const std::vector<size_t> &i, const SizeType &n_bins, const std::vector<T> &lb, const std::vector<T> &ub, const FlagType &logarithmic, const bool option_lt_ub = false)
{
  return histogram_lower_bound(i+(size_t)1, n_bins, lb, ub, logarithmic, option_lt_ub);
}
template <typename T, typename SizeType, typename FlagType>
inline
std::pair<std::vector<T>, std::vector<T> > histogram_range(const std::vector<size_t> &i, const SizeType &n_bins, const std::vector<T> &lb, const std::vector<T> &ub, const FlagType &logarithmic, const bool option_lt_ub = false)
{
  return std::make_pair(histogram_lower_bound(i, n_bins, lb, ub, logarithmic, option_lt_ub),
			histogram_upper_bound(i, n_bins, lb, ub, logarithmic, option_lt_ub));
}
std::map<std::vector<size_t>, size_t>::const_iterator histogram_minimum_element(const std::map<std::vector<size_t>, size_t> &heights)
{
  if (heights.begin() == heights.end()) return heights.end();
  std::map<std::vector<size_t>, size_t>::const_iterator it(heights.begin()), rv(it++);
  for (; it!=heights.end(); ++it)
    if (it->second < rv->second)
      rv = it;
  return rv;
}
inline
std::vector<size_t> histogram_minimum_height_index(const std::map<std::vector<size_t>, size_t> &heights)
{
  return histogram_minimum_element(heights)->first;
}
inline
size_t histogram_minimum_height(const std::map<std::vector<size_t>, size_t> &heights)
{
  return histogram_minimum_element(heights)->second;
}
std::map<std::vector<size_t>, size_t>::const_iterator histogram_maximum_element(const std::map<std::vector<size_t>, size_t> &heights)
{
  if (heights.begin() == heights.end()) return heights.end();
  std::map<std::vector<size_t>, size_t>::const_iterator it(heights.begin()), rv(it++);
  for (; it!=heights.end(); ++it)
    if (rv->second < it->second)
      rv = it;
  return rv;
}
inline
std::vector<size_t> histogram_maximum_height_index(const std::map<std::vector<size_t>, size_t> &heights)
{
  return histogram_maximum_element(heights)->first;
}
inline
size_t histogram_maximum_height(const std::map<std::vector<size_t>, size_t> &heights)
{
  return histogram_maximum_element(heights)->second;
}
template <typename T, typename SizeType, typename FlagType>
std::vector<T> histogram_mean(const std::map<std::vector<size_t>, size_t> &heights, const SizeType &n_bins, const std::vector<T> &lb, const std::vector<T> &ub, const FlagType &logarithmic)
{
  std::vector<INTERNAL_VALUE_TYPE> w(lb.size()), v(lb.size());
  for (std::map<std::vector<size_t>, size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    //if (!(it->second > 0)) continue;
    const std::pair<std::vector<T>, std::vector<T> > range(histogram_range(it->first, n_bins, lb, ub, logarithmic));
#if defined(HISTOGRAM_STATISTICS_NON_LOGARITHMIC_INTERPOLATION_GSL_COMPATIBLE)
    const std::vector<INTERNAL_VALUE_TYPE> p(container_cast<std::vector<INTERNAL_VALUE_TYPE> >((range.first + range.second)/2));
#else
    const std::vector<INTERNAL_VALUE_TYPE> p(operator_conditional(!logarithmic, container_cast<std::vector<INTERNAL_VALUE_TYPE> >(range.first + range.second)/2, container_cast<std::vector<INTERNAL_VALUE_TYPE> >(exp((log(range.first) + log(range.second))/2))));
#endif
    w += it->second;
    v += (p - v)*((INTERNAL_VALUE_TYPE)it->second/w);
  }
  return container_cast<std::vector<T> >(v);
}
template <typename T, typename SizeType, typename FlagType>
std::vector<T> histogram_variance(const std::map<std::vector<size_t>, size_t> &heights, const SizeType &n_bins, const std::vector<T> &lb, const std::vector<T> &ub, const FlagType &logarithmic)
{
  std::vector<INTERNAL_VALUE_TYPE> mu(container_cast<std::vector<INTERNAL_VALUE_TYPE> >(histogram_mean(heights, n_bins, lb, ub, logarithmic))), w(lb.size()), v(lb.size());
  for (std::map<std::vector<size_t>, size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    //if (!(it->second > 0)) continue;
    const std::pair<std::vector<T>, std::vector<T> > range(histogram_range(it->first, n_bins, lb, ub, logarithmic));
#if defined(HISTOGRAM_STATISTICS_NON_LOGARITHMIC_INTERPOLATION_GSL_COMPATIBLE)
    const std::vector<INTERNAL_VALUE_TYPE> p(container_cast<std::vector<INTERNAL_VALUE_TYPE> >((range.first + range.second)/2)), delta(p - mu);
#else
    const std::vector<INTERNAL_VALUE_TYPE> p(operator_conditional(!logarithmic, container_cast<std::vector<INTERNAL_VALUE_TYPE> >((range.first + range.second)/2), container_cast<std::vector<INTERNAL_VALUE_TYPE> >(exp((log(range.first) + log(range.second))/2)))), delta(p - mu);
#endif
    w += it->second;
    v += (delta*delta - v)*((INTERNAL_VALUE_TYPE)it->second/w);
  }
  return container_cast<std::vector<T> >(v);
}
template <typename T, typename SizeType, typename FlagType>
std::vector<T> histogram_standard_deviation(const std::map<std::vector<size_t>, size_t> &heights, const SizeType &n_bins, const std::vector<T> &lb, const std::vector<T> &ub, const FlagType &logarithmic)
{
  return sqrt(histogram_variance(heights, n_bins, lb, ub, logarithmic));
}
template <typename T, typename SizeType, typename FlagType>
T histogram_covariance(const std::map<std::vector<size_t>, size_t> &heights, const SizeType &n_bins, const std::vector<T> &lb, const std::vector<T> &ub, const FlagType &logarithmic)
{
  std::vector<INTERNAL_VALUE_TYPE> mu(container_cast<std::vector<INTERNAL_VALUE_TYPE> >(histogram_mean(heights, n_bins, lb, ub, logarithmic)));
  INTERNAL_VALUE_TYPE w(0), v(0);
  for (std::map<std::vector<size_t>, size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    //if (!(it->second > 0)) continue;
    const std::pair<std::vector<T>, std::vector<T> > range(histogram_range(it->first, n_bins, lb, ub, logarithmic));
#if defined(HISTOGRAM_STATISTICS_NON_LOGARITHMIC_INTERPOLATION_GSL_COMPATIBLE)
    const std::vector<INTERNAL_VALUE_TYPE> p(container_cast<std::vector<INTERNAL_VALUE_TYPE> >((range.first + range.second)/2)), delta(p - mu);
#else
    const std::vector<INTERNAL_VALUE_TYPE> p(operator_conditional(!logarithmic, container_cast<std::vector<INTERNAL_VALUE_TYPE> >((range.first + range.second)/2), container_cast<std::vector<INTERNAL_VALUE_TYPE> >(exp((log(range.first) + log(range.second))/2)))), delta(p - mu);
#endif
    w += it->second;
    v += (prod(delta) - v)*((INTERNAL_VALUE_TYPE)it->second/w);
  }
  return v;
}

template <typename T, typename SizeType>
inline
T histogram_volume_size(const size_t n_dimension, const SizeType &n_bins)
{
  (void)n_dimension;
  return prod(container_cast<std::vector<T> >(n_bins));
}
template <typename T>
inline
T histogram_volume_size(const size_t n_dimension, const size_t n_bins)
{
  return pow(T(n_bins), T(n_dimension));
}

#if __cplusplus < 201103L
template <typename T, typename U>
struct map_value_less {
  bool operator()(const typename std::map<T, U>::value_type &a, const typename std::map<T, U>::value_type::second_type &b)
  { return a.second < b; }
};
#endif

#if defined(HISTOGRAM_CUMULATIVE_DISTRIBUTION_GSL_COMPATIBLE)
template <typename T, typename SizeType>
std::map<std::vector<size_t>, T> histogram_cumulative_distribution(const std::map<std::vector<size_t>, size_t> &heights, const SizeType &n_bins)
{
  std::map<std::vector<size_t>, T> cd;
  T mu(0), sigma(0), volume_size(histogram_volume_size<T>(heights.begin()->first.size(), n_bins));
  for (std::map<std::vector<size_t>, size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    //if (!it->second) continue;
    mu += (it->second - mu)/(std::distance(heights.begin(), it) + 1);
  }
  mu *= heights.size()/volume_size;
  for (std::map<std::vector<size_t>, size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    //if (!it->second) { cd[it->first] = sigma; continue; }
    sigma += it->second/mu/volume_size;
    cd[it->first] = sigma;
  }
  return cd;
}
#else
template <typename T, typename SizeType>
std::map<std::vector<size_t>, T> histogram_cumulative_distribution(const std::map<std::vector<size_t>, size_t> &heights, const SizeType &n_bins)
{
  (void)n_bins;
  std::map<std::vector<size_t>, T> cd;
  T mu(0), sigma(0);
  size_t c(0);
  for (std::map<std::vector<size_t>, size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    //if (!it->second) continue;
    mu += (it->second - mu)/++c;
  }
  for (std::map<std::vector<size_t>, size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    //if (!it->second) { cd[it->first] = sigma; continue; }
    sigma += it->second/mu/c;
    cd[it->first] = sigma;
  }
  return cd;
}
#endif
template <typename T, typename SizeType, typename FlagType>
std::vector<T> histogram_cumulative_distribution_pseudo(std::map<std::vector<size_t>, T> &cd, const std::vector<T> &r,
							const SizeType &n_bins, const std::vector<T> &lb, const std::vector<T> &ub, const FlagType &logarithmic, const bool option_lt_ub = false)
{
  const typename std::map<std::vector<size_t>, T>::const_iterator it =
    lower_bound(cd.begin(), cd.end(), r[0],
#if __cplusplus < 201103L
		map_value_less<std::vector<size_t>, T>()
#else
		[](const typename std::map<std::vector<size_t>, T>::value_type &a, const typename std::map<std::vector<size_t>, T>::value_type::second_type &b) ->
		bool { return a.second < b; }
#endif
		);
  const std::pair<std::vector<T>, std::vector<T> > range(histogram_range(it->first, n_bins, lb, ub, logarithmic, option_lt_ub));
  std::vector<T> delta(r);
  delta[0] = (r[0] - (it==cd.begin() ? T(0) : cd[prev(it)->first]))/(cd[it->first] - (it==cd.begin() ? T(0) : cd[prev(it)->first]));
#if defined(HISTOGRAM_PSEUDO_DENSITY_NON_LOGARITHMIC_INTERPOLATION_GSL_COMPATIBLE)
  return
    range.first + delta*(range.second - range.first);
#else
  return !logarithmic ?
    range.first + delta*(range.second - range.first) :
    range.first*exp(delta*(log(range.second) - log(range.first)));
#endif
}

#if defined(HISTOGRAM_CUMULATIVE_DISTRIBUTION_GSL_COMPATIBLE)
template <typename T, typename SizeType>
std::pair<std::map<std::vector<size_t>, T>, std::map<std::vector<size_t>, T> > histogram_probability_cumulative_distribution(const std::map<std::vector<size_t>, size_t> &heights, const SizeType &n_bins)
{
  (void)n_bins;
  std::map<std::vector<size_t>, T> pd, cd;
  T mu(0), sigma(0), volume_size(histogram_volume_size<T>(heights.begin()->first.size(), n_bins));
  for (std::map<std::vector<size_t>, size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    //if (!it->second) continue;
    mu += (it->second - mu)/(std::distance(heights.begin(), it) + 1);
  }
  mu *= heights.size()/volume_size;
  for (std::map<std::vector<size_t>, size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    //if (!it->second) { pd[it->first] = T(0); cd[it->first] = sigma; continue; }
    pd[it->first] = it->second/mu/volume_size;
    sigma += pd[it->first];
    cd[it->first] = sigma;
  }
  return std::make_pair(pd, cd);
}
#else
template <typename T, typename SizeType>
std::pair<std::map<std::vector<size_t>, T>, std::map<std::vector<size_t>, T> > histogram_probability_cumulative_distribution(const std::map<std::vector<size_t>, size_t> &heights, const SizeType &n_bins)
{
  (void)n_bins;
  std::map<std::vector<size_t>, T> pd, cd;
  T mu(0), sigma(0);
  size_t c(0);
  for (std::map<std::vector<size_t>, size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    //if (!it->second) continue;
    mu += (it->second - mu)/++c;
  }
  for (std::map<std::vector<size_t>, size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    //if (!it->second) { pd[it->first] = T(0); cd[it->first] = sigma; continue; }
    pd[it->first] = it->second/mu/c;
    sigma += pd[it->first];
    cd[it->first] = sigma;
  }
  return std::make_pair(pd, cd);
}
#endif
template <typename T, typename SizeType, typename FlagType>
std::vector<T> histogram_probability_cumulative_distribution_pseudo(std::pair<std::map<std::vector<size_t>, T>, std::map<std::vector<size_t>, T> > &pd_cd, const std::vector<T> &r,
								    const SizeType &n_bins, const std::vector<T> &lb, const std::vector<T> &ub, const FlagType &logarithmic, const bool option_lt_ub = false)
{
  const typename std::map<std::vector<size_t>, T>::const_iterator it =
    lower_bound(pd_cd.second.begin(), pd_cd.second.end(), r[0],
#if __cplusplus < 201103L
		map_value_less<std::vector<size_t>, T>()
#else
		[](const typename std::map<std::vector<size_t>, T>::value_type &a, const typename std::map<std::vector<size_t>, T>::value_type::second_type &b) ->
		bool { return a.second < b; }
#endif
		);
  const std::pair<std::vector<T>, std::vector<T> > range(histogram_range(it->first, n_bins, lb, ub, logarithmic, option_lt_ub));
  std::vector<T> delta(r);
#if defined(HISTOGRAM_PSEUDO_DENSITY_GSL_COMPATIBLE)
  delta[0] = (r[0] - (it==pd_cd.second.begin() ? T(0) : pd_cd.second[prev(it)->first]))/(pd_cd.second[it->first] - (it==pd_cd.second.begin() ? T(0) : pd_cd.second[prev(it)->first]));
#if defined(HISTOGRAM_PSEUDO_DENSITY_NON_LOGARITHMIC_INTERPOLATION_GSL_COMPATIBLE)
  return
    range.first + delta*(range.second - range.first);
#else
  return !logarithmic ?
    range.first + delta*(range.second - range.first) :
    range.first*exp(delta*(log(range.second) - log(range.first)));
#endif
#else
  delta[0] = (pd_cd.second[it->first] - r[0])/pd_cd.first[it->first];
#if defined(HISTOGRAM_PSEUDO_DENSITY_NON_LOGARITHMIC_INTERPOLATION_GSL_COMPATIBLE)
  return
    delta*range.first + (1 - delta)*range.second;
#else
  return !logarithmic ?
    delta*range.first + (1 - delta)*range.second :
    exp(delta*log(range.first) + (1 - delta)*log(range.second));
#endif
#endif
}

#endif
