/*
  multidimensional_histograms.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <vector>
#include <list>
#include <functional>
#if __cplusplus < 201103L || defined(USE_BOOST_FUNCTION)
#include <boost/function.hpp>
#endif
#include <cstdlib>      // random(), etc
#include <cassert>
#if __cplusplus < 201103L
#include <boost/utility.hpp>
#endif
#include "statistics.hpp"
#include "multidimensional_histograms.hpp"
#include "multidimensional_histograms_online.hpp"

#define numberof(a) (sizeof(a)/sizeof(a[0]))

//#define USE_LOGARITHMIC_HISTOGRAM

std::vector<double> henon_map_initial_values()
{
  std::vector<double> v(2);
  v[0] = .2; v[1] = .1;
  return v;
}
std::vector<double> henon_map(const std::vector<double> &v0, const double a = 1.4, const double b = .3)
{
  std::vector<double> v1(v0.size());
#define x0 v0[0]
#define y0 v0[1]
#define x1 v1[0]
#define y1 v1[1]
  x1 = y0 + 1 - a*x0*x0;
  y1 = b*x0;
#undef x0
#undef y0
#undef x1
#undef y1
  return v1;
}

int main()
{
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
  bool logarithmic(false);
#else
  bool logarithmic(true);
#endif
  const double maximum_permissible_error(1e-2);
  const char *scalar_result_names[] = {
    "histogram min_value",
    "histogram max_value",
    "histogram total",
    "histogram sum",
    "histogram volume size",
    "histogram non-zero volume size",
    "histogram covariance",
  };
  const char *result_names[] = {
    "minimum value",
    "maximum value",
    "histogram lower bound",
    "histogram upper bound",
    "histogram min_element",
    "histogram max_element",
    "histogram mean",
    "histogram sigma",
  };
  std::vector<std::vector<double> > v(100000);
  const size_t n_dimension(2), n_bins(v.size()/1000), n_prior_points(v.size());
  multidimensional_histogram_online<double, size_t, bool> pseudo_dense_histogram_online(n_bins, n_dimension, n_prior_points, logarithmic);

  std::vector<double> w(henon_map_initial_values());
  for (size_t i=0; i<v.size(); ++i) {
    w = henon_map(w);
    v[i] = w;
  }
#if !defined(USE_LOGARITHMIC_HISTOGRAM)
#else
  std::vector<double> bias(-min_values(v.begin(), v.end()));
  for (size_t i=0; i<v.size(); ++i)
    v[i] += bias;
#endif
  std::vector<double> result_range[] = {
    !logarithmic ? min_values(v.begin(), v.end()) : positive_min_values(v.begin(), v.end()),
    max_values(v.begin(), v.end()),
  };
  std::map<std::vector<size_t>, size_t> histogram_result;
  size_t histogram_result_summation(histogram(histogram_result, v.begin(), v.end(), n_bins, result_range[0], result_range[1], logarithmic));
  double scalar_results[] = {
    (double)histogram_minimum_height(histogram_result),
    (double)histogram_maximum_height(histogram_result),
    (double)std::distance(v.begin(), v.end()),
#if __cplusplus < 201103L
    (double)histogram_result_summation,
#else
    (double)std::accumulate(histogram_result.begin(), histogram_result.end(), size_t(0),
			    [](const decltype(histogram_result)::value_type::second_type &a, const decltype(histogram_result)::value_type &b) ->
			    decltype(histogram_result)::value_type::second_type { return a + b.second; }),
#endif
    histogram_volume_size<double>(n_dimension, n_bins),
    (double)histogram_result.size(),
    histogram_covariance(histogram_result, n_bins, result_range[0], result_range[1], logarithmic),
  };
  std::vector<double> results[] = {
    result_range[0],
    result_range[1],
    histogram_lower_bound(std::vector<size_t>(n_dimension, 0), n_bins, result_range[0], result_range[1], logarithmic),
    histogram_upper_bound(std::vector<size_t>(n_dimension, n_bins-1), n_bins, result_range[0], result_range[1], logarithmic),
    container_cast<std::vector<double> >(histogram_minimum_height_index(histogram_result)),
    container_cast<std::vector<double> >(histogram_maximum_height_index(histogram_result)),
    histogram_mean(histogram_result, n_bins, result_range[0], result_range[1], logarithmic),
    histogram_standard_deviation(histogram_result, n_bins, result_range[0], result_range[1], logarithmic),
  };

#if !0 /* worse implement */
  std::map<std::vector<size_t>, double> pd_histogram(histogram_cumulative_distribution<double>(histogram_result, n_bins));
#else
  std::pair<std::map<std::vector<size_t>, double>, std::map<std::vector<size_t>, double> > pd_histogram(histogram_probability_cumulative_distribution<double>(histogram_result, n_bins));
#endif

  srandomdev();
  for (size_t i=0; i<v.size(); ++i) {
    std::vector<double> v(n_dimension);
    for (size_t j=0; j<n_dimension; ++j) {
      double x((double)random()/((1L<<31)-1));
      v[j] = x;
    }
#if !0 /* worse implement */
    pseudo_dense_histogram_online.add(histogram_cumulative_distribution_pseudo(pd_histogram, v,
									       n_bins, result_range[0], result_range[1], logarithmic));
#else
    pseudo_dense_histogram_online.add(histogram_probability_cumulative_distribution_pseudo(pd_histogram, v,
											   n_bins, result_range[0], result_range[1], logarithmic));
#endif
  }

  double pd_scalar_results[] = {
    (double)pseudo_dense_histogram_online.get_minimum_height(),
    (double)pseudo_dense_histogram_online.get_maximum_height(),
    (double)pseudo_dense_histogram_online.get_count(),
    (double)pseudo_dense_histogram_online.get_heights_summation(),
    histogram_volume_size<double>(n_dimension, n_bins),
    pseudo_dense_histogram_online.get_heights().size(),
    pseudo_dense_histogram_online.get_heights_covariance(),
  };
  std::vector<double> pd_results[] = {
    !logarithmic ? pseudo_dense_histogram_online.get_minimum_value() : pseudo_dense_histogram_online.get_positive_minimum_value(),
    pseudo_dense_histogram_online.get_maximum_value(),
    pseudo_dense_histogram_online.get_lower_bound(std::vector<size_t>(n_dimension, 0)),
    pseudo_dense_histogram_online.get_upper_bound(std::vector<size_t>(n_dimension, n_bins-1)),
    container_cast<std::vector<double> >(pseudo_dense_histogram_online.get_minimum_height_index()),
    container_cast<std::vector<double> >(pseudo_dense_histogram_online.get_maximum_height_index()),
    pseudo_dense_histogram_online.get_heights_mean(),
    pseudo_dense_histogram_online.get_heights_standard_deviation(),
  };
  for (size_t j=0; j<numberof(pd_scalar_results); ++j) {
    double error(sqrt(pow(pd_scalar_results[j] - scalar_results[j], 2)) / (scalar_results[j] ? fabs(scalar_results[j]) : 1));
    std::cout << scalar_result_names[j] << ":\t" << pd_scalar_results[j] << "\t" << scalar_results[j] << "\t" << error << std::endl;
    if (!(j==0||j==1||j==5||j==6)) assert(error < maximum_permissible_error);
  }
  for (size_t j=0; j<numberof(pd_results); ++j) {
    std::vector<double> error(sqrt(pow(pd_results[j] - results[j], 2)));
    std::cout << result_names[j] << ":" << std::endl;
    for (size_t i=0; i<n_dimension; ++i)
      std::cout << "\t" << i << ":\t" << pd_results[j][i] << "\t" << results[j][i] << "\t" << error[i] << std::endl;
    if (!(j==4||j==5)) assert(*max_element(error.begin(), error.end()) < maximum_permissible_error);
  }

  return 0;
}
